"""Tests for pandas series and dataframe method registration."""

import pandas_flavor as pf
import pandas as pd
from pandas.core.groupby.generic import DataFrameGroupBy


def test_register_dataframe_method():
    """Test register_dataframe_method."""

    @pf.register_dataframe_method
    def dummy_func(df: pd.DataFrame) -> pd.DataFrame:
        """Dummy function.

        Args:
            df: a pandas DataFrame

        Returns:
            df: A pandas DataFrame.
        """
        return df

    df = pd.DataFrame()
    df.dummy_func()


def test_register_series_method():
    """Test register_series_method."""

    @pf.register_series_method
    def dummy_func(s: pd.Series) -> pd.Series:
        """Dummy func.

        Args:
            s: A pandas Series.

        Returns:
            s: A pandas Series.
        """
        return s

    ser = pd.Series()
    ser.dummy_func()


def test_register_groupby_method():
    """Test register_groupby_method."""

    @pf.register_groupby_method
    def dummy_func(by: DataFrameGroupBy) -> DataFrameGroupBy:
        """Dummy func.

        Args:
            by: A DataFrameGroupBy object.

        Returns:
            DataFrameGroupBy.
        """
        return by

    df = pd.DataFrame(
        {
            "Animal": ["Falcon"],
            "Max Speed": [380.0],
        }
    )
    by = df.groupby("Animal")
    by.dummy_func()



"""XArray support for pandas_flavor."""
from xarray import register_dataarray_accessor, register_dataset_accessor
from functools import wraps


def make_accessor_wrapper(method):
    """
    Makes an XArray-compatible accessor to wrap a method to be added to an
    xr.DataArray, xr.Dataset, or both.

    Args:
        method: A method which takes an XArray object and needed parameters.

    Returns:
        The result of calling ``method``.
    """

    class XRAccessor:
        """XArray accessor for a method."""

        def __init__(self, xr_obj):
            """Initialize the accessor.

            Args:
                xr_obj: The XArray object to which the accessor is attached.
            """
            self._xr_obj = xr_obj

        @wraps(method)
        def __call__(self, *args, **kwargs):
            """Call the method.

            Args:
                *args: Positional arguments to pass to the method.
                **kwargs: Keyword arguments to pass to the method.

            Returns:
                The result of calling ``method``.
            """

            return method(self._xr_obj, *args, **kwargs)

    return XRAccessor


def register_xarray_dataarray_method(method: callable):
    """Register a method on an XArray DataArray object.

    Args:
        method: A method which takes an XArray object and needed parameters.

    Returns:
        The method.
    """
    accessor_wrapper = make_accessor_wrapper(method)
    register_dataarray_accessor(method.__name__)(accessor_wrapper)

    return method


def register_xarray_dataset_method(method: callable):
    """Register a method on an XArray Dataset object.

    Args:
        method: A method which takes an XArray object and needed parameters.

    Returns:
        The method.
    """

    accessor_wrapper = make_accessor_wrapper(method)
    register_dataset_accessor(method.__name__)(accessor_wrapper)

    return method



"""Register functions as methods of Pandas DataFrame and Series."""

from __future__ import annotations

import warnings
from functools import wraps

from pandas.core.groupby.generic import DataFrameGroupBy
from pandas.util._exceptions import find_stack_level
from pandas.api.extensions import (
    register_series_accessor,
    register_dataframe_accessor,
)
import inspect

method_call_ctx_factory = None


def handle_pandas_extension_call(method, method_signature, obj, args, kwargs):
    """Handle pandas extension call.

    This function is called when the user calls
    a pandas DataFrame object's method.
    The pandas extension mechanism passes args and kwargs
    of the original method call as it is applied to obj.

    Our implementation uses the global variable `method_call_ctx_factory`.

    `method_call_ctx_factory` can be either None or an abstract class.

    When `method_call_ctx_factory` is None,
    the implementation calls the registered method
    with unmodified args and kwargs and returns underlying method result.

    When `method_call_ctx_factory` is not None,
    `method_call_ctx_factory` is expected to refer to
    the function to create the context object.
    The context object will be used to process
    inputs and outputs of `method` calls.
    It is also possible that
    the context object method `handle_start_method_call`
    will modify original args and kwargs before `method` call.

    `method_call_ctx_factory` is a function
    that should have the following signature:

     `f(method_name: str, args: list, kwargs: dict) -> MethodCallCtx`


    MethodCallCtx is an abstract class:
    class MethodCallCtx(abc.ABC):

        @abstractmethod
        def __enter__(self) -> None:
            raise NotImplemented

        @abstractmethod
        def __exit__(self, exc_type, exc_value, traceback) -> None:
            raise NotImplemented

        @abstractmethod
        def handle_start_method_call(self, method_name: str, method_signature: inspect.Signature, method_args: list, method_kwargs: dict) -> tuple(list, dict):
            raise NotImplemented

        @abstractmethod
        def handle_end_method_call(self, ret: object) -> None:
            raise NotImplemented


    Args:
        method (callable): method object as registered by decorator
            register_dataframe_method (or register_series_method)
        method_signature: signature of method as returned by inspect.signature
        obj: Dataframe or Series
        args: The arguments to pass to the registered method.
        kwargs: The keyword arguments to pass to the registered method.

    Returns:
        object`: The result of calling of the method.
    """  # noqa: E501

    global method_call_ctx_factory
    with method_call_ctx_factory(
        method.__name__, args, kwargs
    ) as method_call_ctx:
        if method_call_ctx is None:  # nullcontext __enter__ returns None
            ret = method(obj, *args, **kwargs)
        else:
            all_args = tuple([obj] + list(args))
            (
                new_args,
                new_kwargs,
            ) = method_call_ctx.handle_start_method_call(
                method.__name__, method_signature, all_args, kwargs
            )
            args = new_args[1:]
            kwargs = new_kwargs

            ret = method(obj, *args, **kwargs)

            method_call_ctx.handle_end_method_call(ret)

        return ret


def register_dataframe_method(method):
    """Register a function as a method attached to the Pandas DataFrame.

    Example:

        @register_dataframe_method
        def print_column(df, col):
            '''Print the dataframe column given'''
            print(df[col])

    Args:
        method (callable): callable to register as a dataframe method.

    Returns:
        callable: The original method.
    """

    method_signature = inspect.signature(method)

    def inner(*args, **kwargs):
        """Inner function to register the method.

        This function is called when the user
        decorates a function with register_dataframe_method.

        Args:
            *args: The arguments to pass to the registered method.
            **kwargs: The keyword arguments to pass to the registered method.

        Returns:
            method: The original method.
        """

        class AccessorMethod(object):
            """DataFrame Accessor method class."""

            def __init__(self, pandas_obj):
                """Initialize the accessor method class.

                Args:
                    pandas_obj (pandas.DataFrame): The pandas DataFrame object.
                """
                self._obj = pandas_obj

            @wraps(method)
            def __call__(self, *args, **kwargs):
                """Call the accessor method.

                Args:
                    *args: The arguments to pass to the registered method.
                    **kwargs: The keyword arguments to pass
                        to the registered method.

                Returns:
                    object: The result of calling of the method.
                """
                global method_call_ctx_factory
                if method_call_ctx_factory is None:
                    return method(self._obj, *args, **kwargs)

                return handle_pandas_extension_call(
                    method, method_signature, self._obj, args, kwargs
                )

        register_dataframe_accessor(method.__name__)(AccessorMethod)

        return method

    return inner()


def register_series_method(method):
    """Register a function as a method attached to the Pandas Series.

    Args:
        method (callable): callable to register as a series method.

    Returns:
        callable: The original method.
    """

    method_signature = inspect.signature(method)

    def inner(*args, **kwargs):
        """Inner function to register the method.

        Args:
            *args: The arguments to pass to the registered method.
            **kwargs: The keyword arguments to pass to the registered method.

        Returns:
            method: The original method.
        """

        class AccessorMethod(object):
            """Series Accessor method class."""

            __doc__ = method.__doc__

            def __init__(self, pandas_obj):
                """Initialize the accessor method class.

                Args:
                    pandas_obj (pandas.Series): The pandas Series object.
                """
                self._obj = pandas_obj

            @wraps(method)
            def __call__(self, *args, **kwargs):
                """Call the accessor method.

                Args:
                    *args: The arguments to pass to the registered method.
                    **kwargs: The keyword arguments to pass
                        to the registered method.

                Returns:
                    object: The result of calling of the method.
                """
                global method_call_ctx_factory
                if method_call_ctx_factory is None:
                    return method(self._obj, *args, **kwargs)

                return handle_pandas_extension_call(
                    method, method_signature, self._obj, args, kwargs
                )

        register_series_accessor(method.__name__)(AccessorMethod)

        return method

    return inner()


# variant of pandas' accessor

# copied from pandas' accessor file - pandas/pandas/core/accessor.py
"""

accessor.py contains base classes for implementing accessor properties
that can be mixed into or pinned onto other pandas classes.

"""


class CachedAccessor:
    """
    Custom property-like object.

    A descriptor for caching accessors.

    Parameters
    ----------
    name : str
        Namespace that will be accessed under, e.g. ``df.foo``.
    accessor : DataFrameGroupBy
        Class with the extension methods.

    Notes
    -----
    For accessor, The class's __init__ method assumes that one of
    ``Series``, ``DataFrame`` or ``Index`` as the
    single argument ``data``.
    """

    def __init__(self, name: str, accessor: DataFrameGroupBy) -> None:
        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            # we're accessing the attribute of the class, i.e., Dataset.geo
            return self._accessor
        accessor_obj = self._accessor(obj)
        # Replace the property with the accessor object. Inspired by:
        # https://www.pydanny.com/cached-property.html
        # We need to use object.__setattr__ because we overwrite __setattr__ on
        # NDFrame
        object.__setattr__(obj, self._name, accessor_obj)
        return accessor_obj


def _register_accessor(name: str, cls: DataFrameGroupBy):
    """
    Register a custom accessor on a DataFrameGroupBy object.

    Args:
        name : str
            Name under which the accessor should be registered.
            A warning is issued
            if this name conflicts with a preexisting attribute.
        cls: DataFrameGroupBy

    Returns:
        A class decorator.
    """

    def decorator(accessor):
        if hasattr(cls, name):
            warnings.warn(
                f"registration of accessor {repr(accessor)} under name "
                f"{repr(name)} for type {repr(cls)} "
                "is overriding a preexisting "
                f"attribute with the same name.",
                UserWarning,
                stacklevel=find_stack_level(),
            )
        setattr(cls, name, CachedAccessor(name, accessor))
        if not hasattr(cls, "_accessors"):
            cls._accessors = set()
        cls._accessors.add(name)
        return accessor

    return decorator


def register_groupby_accessor(name: str):
    return _register_accessor(name, DataFrameGroupBy)


def register_groupby_method(method):
    """Register a function as a method attached to the pandas DataFrameGroupBy.

    Example:
        >>> @register_groupby_method # doctest: +SKIP
        >>> def print_column(grp, col): # doctest: +SKIP
        ...    '''Print the dataframe column given''' # doctest: +SKIP
        ...    print(grp[col]) # doctest: +SKIP

    !!! info "New in version 0.7.0"

    Args:
        method: Function to be registered as a method
            on the DataFrameGroupBy object.

    Returns:
        callable: The original method.
    """
    method_signature = inspect.signature(method)

    def inner(*args: tuple, **kwargs: dict):
        """Inner function to register the method.

        This function is called when the user
        decorates a function with register_groupby_method.

        Args:
            *args: The arguments to pass to the registered method.
            **kwargs: The keyword arguments to pass to the registered method.

        Returns:
            method: The original method.
        """

        class AccessorMethod(object):
            """DataFrameGroupBy Accessor method class."""

            __doc__ = method.__doc__

            def __init__(self, obj):
                """Initialize the accessor method class.

                Args:
                    obj: The pandas DataFrameGroupBy object.
                """
                self._obj = obj

            @wraps(method)
            def __call__(self, *args, **kwargs):
                """Call the accessor method.

                Args:
                    *args: The arguments to pass to the registered method.
                    **kwargs: The keyword arguments to pass
                        to the registered method.

                Returns:
                    object: The result of calling of the method.
                """
                global method_call_ctx_factory
                if method_call_ctx_factory is None:
                    return method(self._obj, *args, **kwargs)

                return handle_pandas_extension_call(
                    method, method_signature, self._obj, args, kwargs
                )

        register_groupby_accessor(method.__name__)(AccessorMethod)
        return method

    return inner()



"""Tracing extension demo."""
import time

import pandas as pd

import pandas_flavor as pf


@pf.register_dataframe_method
def my_method(df: pd.DataFrame) -> pd.DataFrame:
    """Transpose the DataFrame.

    Args:
        df: The DataFrame to transpose.

    Returns:
        The transposed DataFrame.
    """
    print("my_method called")
    return df.transpose()


@pf.register_dataframe_method
def another_method(df: pd.DataFrame, new_col_d) -> pd.DataFrame:
    """Adds a new column to the DataFrame.

    Args:
        df (DataFrame): The DataFrame to add the column to.
        new_col_d (dict): A dictionary of column names and values.

    Returns:
        DataFrame: The DataFrame with the new column.
    """
    print("another_method called")
    for col, v in new_col_d.items():
        df[col] = v
    return df


class tracer:
    """A simple tracer for method calls."""

    @staticmethod
    def create_tracer(*args):
        """Creates a tracer.

        Args:
            *args: Variable length argument list for the tracer.

        Returns:
            The created tracer.
        """
        return tracer()

    def __init__(self):
        """Initialize the tracer."""
        self.method_name = None
        self.start_ts = None
        self.end_ts = None

    def __enter__(self):
        """Enter the tracer.

        Returns:
            The tracer.
        """
        return self

    def handle_start_method_call(
        self, method_name, method_signature, method_args, method_kwagrs
    ):
        """Handle the start of a method call.

        Args:
            method_name: The name of the method.
            method_signature: The signature of the method.
            method_args: The arguments of the method.
            method_kwagrs: The keyword arguments of the method.

        Returns:
            The arguments and keyword arguments of the method.
        """
        self.method_name = method_name
        self.start_ts = time.time()
        return method_args, method_kwagrs

    def handle_end_method_call(self, ret):
        """Handle the end of a method call.

        Args:
            ret: The return value of the method.
        """
        self.end_ts = time.time()

    def __exit__(self, exc_type, value, traceback):
        """Exit the tracer.

        Args:
            exc_type: The type of the exception.
            value: The value of the exception.
            traceback: The traceback of the exception.
        """
        call_dur = self.end_ts - self.start_ts
        print(f"method {self.method_name} took {call_dur} secs to execute")


pf.register.method_call_ctx_factory = tracer.create_tracer
s_df = pd.DataFrame([[i + j for i in range(10)] for j in range(10)])
res_df = s_df.my_method().another_method({"new_col": "new value"})
print(res_df)



"""Top-level API for pandas-flavor."""
from .register import (
    register_dataframe_accessor,
    register_dataframe_method,
    register_series_accessor,
    register_series_method,
    register_groupby_accessor,
    register_groupby_method,
)
from .xarray import (
    register_xarray_dataarray_method,
    register_xarray_dataset_method,
)

__all__ = [
    "register_series_method",
    "register_series_accessor",
    "register_dataframe_method",
    "register_dataframe_accessor",
    "register_groupby_accessor",
    "register_groupby_method",
    "register_xarray_dataarray_method",
    "register_xarray_dataset_method",
]
